; defined in linker script
.extern _istack
.extern _ustack
.extern _sdata
.extern _sidata
.extern _edata

.extern __preinit_array_start
.extern __preinit_array_end
.extern __init_array_start
.extern __init_array_end
.extern __fini_array_start
.extern __fini_array_end

; defined in source file
.extern _HardwareSetup
.extern _main

.text
.global _start
.type _start, %function
_start:
	mvtc #_istack, isp ; write address
	mvtc #_ustack, usp
	mvtc #_srvectors, intb ; interrupt table register
	mvtc #100h, fpsw ; floating point

	; copy .text to ram
	mov #_sdata, r1
	mov #_sidata, r2
	mov #(_edata - _sdata), r3
	smovf ; string move

	; clear all RAM area
	mov #_sbss, r1
	mov #0, r2
	mov #(_ebss - _sbss), r3
	sstr.b ; memset 0(string store)

	mov.l #((1 << 16) | (1 << 17)), r1 ; allow interrupts, use USP for stack

	; goto user-mode
	or #(1 << 20), r1
	mvtc r1, bpsw
	mvtc #execute, bpc
	rtfi

.type constructor, %function
constructor:
preinit:
	mov #__preinit_array_start, r1
preinit_loop:
	cmp #__preinit_array_end, r1 ; evaluate r1 - __pre...
	bge init
	mov.l [r1+], r2
	jsr r2
	bra preinit_loop
init:
	mov #__init_array_start, r1
init_loop:
	cmp #__init_array_end, r1
	bge constructor_end
	mov.l [r1+], r2
	jsr r2
	bra init_loop
constructor_end:
	rts

.type destructor, %function
destructor:
fini:
	mov #__fini_array_start, r1
fini_loop:
	cmp #__fini_array_end, r1
	bge destructor_end
	mov.l [r1+], r2
	jsr r2
	bra fini_loop
destructor_end:
	rts

.type execute, %function
execute:
	bsr _HardwareSetup
	bsr constructor
	bsr _main
	bsr destructor
	bra .
